import logging
import ipyvuetify as v
import ipywidgets as w
import plotly.colors as pc
import plotly.express as px
import plotly.graph_objs as go
import plotly.io as pio
import base64
from typing import Optional

logger = logging.getLogger(__name__)


def set_default_template():
    pio.templates.default = "plotly_white"
    pio.templates[pio.templates.default].layout.colorway = px.colors.sequential.Viridis  # type: ignore


def create_widget(
    fig: go.Figure,
    setcolors: bool,
    png: bool,
    height: Optional[int] = None,
    width: Optional[int] = None,
) -> w.Widget:
    if setcolors:
        default_color_scale = pc.DEFAULT_PLOTLY_COLORS
        numcolors = len(default_color_scale)

        line_dash_sequence = ["solid", "dot", "dash"]
        for i, trace in enumerate(fig.data):
            if hasattr(trace, "line"):
                line_dash_index = (i // numcolors) % len(line_dash_sequence)
                trace.line.dash = line_dash_sequence[line_dash_index]  # type: ignore

    if height is not None:
        fig.layout.height = height
    if width is not None:
        fig.layout.width = width

    if png:
        raise ValueError(
            "Not supported at this time due to kaleido hanging on some environments."
        )
        import kaleido  # type: ignore  # noqa

        # Kaleido is only required if png generation is needed. This ensures it is installed.
        # App may hang otherwise

        # if width is None:
        #    fig.layout.width = 900
        image_bytes = pio.to_image(fig, format="png")
        encoded_image = base64.b64encode(image_bytes).decode("utf-8")

        w = v.Html(
            tag="img",
            children=[],
            attributes={"src": "data:image/png;base64," + encoded_image},
        )

    else:
        w = go.FigureWidget(fig)

    return w



from .browser_title import BrowserTitle  # noqa: I001

from .appinator import SelectTypes, TabbedUiModel, UiModel, UiPage, UiWidget
from .datagrid import ColMd
from ._version import __version__

__all__ = [
    "__version__",
    "TabbedUiModel",
    "UiModel",
    "UiPage",
    "BrowserTitle",
    "SelectTypes",
    "UiWidget",
    "ColMd",
]



class BrowserTitle:
    title: str
    iconHref: str

    def __init__(self, title: str, iconHref: str = "https://iqmo.com/iqmo-dk.svg"):
        self.title = title
        self.iconHref = iconHref

    def _repr_html_(self):
        return f"""
            <script>
                (()=>{{
                    document.title='{self.title}'
                    let link = document.querySelector("head link[rel~='icon']");
                    if(!link){{
                        link = document.createElement('link');
                        link.rel = 'icon';
                        document.head.appendChild(link);
                    }}
                    link.href = '{self.iconHref}'
                }})()
            </script>
        """



import logging
import json
import pandas as pd
from dataclasses import dataclass

import ipyaggrid as ag
from typing import List, Dict, Literal, Optional
from typing import Any, Callable, Tuple, Union

logger = logging.getLogger(__name__)

# Set this to enable Enterprise features
LICENSE = "community"


def get_license():
    return LICENSE


@dataclass
class ColMd:
    name: str
    label: Optional[str] = None
    format: str = "default"  # percent, decimal, mag_si (convert to K/M/G suffixes)
    pinned: bool = False
    hidden: bool = False
    precision: Optional[int] = None

    def __post_init__(self):
        if not isinstance(self.name, str):
            self.name = str(self.name)
        if self.label is None:
            self.label = self.name.title().replace("_", " ")


def _addcol(
    md: ColMd,
    allcols: Dict[str, Dict[str, object]],
    autodetect: bool = True,
    precision: int = 2,
):
    col = md.name
    if col not in allcols:
        precision = md.precision if md.precision is not None else precision

        if autodetect and md.format == "default":
            if col.endswith("perc"):
                md.format = "perc"
                md.precision = precision

        if md.format.startswith("perc"):
            allcols[col] = {
                "field": col,
                "width": 150,
                "minWidth": 100,
                "hide": md.hidden,
                "cellStyle": {"text-align": "right"},
                "valueFormatter": """
                function (node) {
                    try {
                        if (node === null || node.value === null) {
                            return null;
                        } else {
                            return node.value.toLocaleString(undefined, { style: 'percent', minimumFractionDigits: """
                + str(precision)
                + """ });
                        }
                    } catch (error) {
                        return node.value;
                    }
                }""",
            }
        elif md.format.startswith("dec"):
            formatter = (
                """function (node) {
                    try {
                        if (node === null || node.value === null) {
                            return null;
                        } 
                        else if (node.value == 0) {
                            return node.value;
                        }
                        else {
                            return node.value.toLocaleString(undefined, { minimumFractionDigits: """
                + str(precision)
                + """, maximumFractionDigits: """
                + str(precision)
                + """  });
                        }
                    } catch (error) {
                        return node.value;
                    }
                }
                """
            )
            allcols[col] = {
                "field": col,
                "width": 150,
                "minWidth": 100,
                "cellStyle": {"text-align": "right"},
                "valueFormatter": formatter,
            }
        elif md.format == "mag_si":
            formatter = (
                """function (node) {
                    try {
                        if (node === null || node.value === null) {
                            return null;
                        } 
                        
                        let result = null;
                        let suffix = null;
                        if (node.value >= 0 && node.value <1000) {
                            result = node.value;
                            suffix = "";
                        }
                        else if (node.value >= 1000 && node.value <1000000) {
                            result = node.value/1000;
                            suffix = "K";
                        }
                        else if (node.value >= 1000000 && node.value <=1000000000) {
                            result = node.value/1000000;
                            suffix = "M";
                        }
                        else {
                            result = node.value/1000000000;
                            suffix = "B";
                        }
                        let result_round = result.toLocaleString(undefined, { minimumFractionDigits: """
                + str(precision)
                + """, maximumFractionDigits: """
                + str(precision)
                + """  }) + suffix;
                            
                        return result_round;                    
                    } catch (error) {
                        console.log(error);
                        return node.value;
                    }
                }
                """
            )
            allcols[col] = {
                "field": col,
                "width": 150,
                "minWidth": 100,
                "cellStyle": {"text-align": "right"},
                "valueFormatter": formatter,
            }
        else:
            allcols[col] = {
                "field": col,
                "width": 150,
                "minWidth": 100,
                "cellStyle": {"text-align": "right"},
                "valueFormatter": f"""function (node) {{
                    try {{
                        if (node === null || node.value === null) {{
                            return null;
                        }} else {{
                            return node.value.toLocaleString(undefined, {{ minimumFractionDigits: {str(precision)}, maximumFractionDigits: {str(precision)}  }});
                        }}
                    }} catch (error) {{
                        return node.value;
                    }}
                }}
                """,
            }

        # pinned = collower in pinned
        if md.pinned:
            allcols[col]["pinned"] = True
        if md.hidden:
            allcols[col]["hide"] = True
        allcols[col]["headerName"] = md.label


class DataGrid(ag.Grid):
    message_handlers: List[Tuple[Callable[[Dict], None], Union[str, None]]]
    current_selection: Optional[Dict] = None
    _select_mode: Optional[Literal["single", "multiple"]]

    def __init__(
        self,
        grid_data: Union[List, pd.DataFrame],
        is_tree: bool = False,
        select_mode: Optional[Literal["single", "multiple"]] = "single",
        events=None,
        pathcol: Optional[str] = "path",
        pathdelim="/",
        col_md: List[ColMd] = [],
        showindex: bool = False,
        js_post_grid: List = [],
        js_pre_grid: List = [],
        grid_options: Dict = {},
        num_toppinned_rows=0,
        flatten_columns=True,
        default_precision=2,
        *argv,
        **kargv,
    ):
        license = get_license()
        # if pathcol is not None:
        #    pathcol = pathcol.title()
        sortcols = False
        if isinstance(grid_data, pd.DataFrame) and flatten_columns:
            input_df = grid_data.copy()
            input_df.columns = [str(col) for col in input_df.columns]
        else:
            input_df = pd.DataFrame(grid_data)

        if showindex:
            input_df = input_df.reset_index()

        allcols = {}

        if select_mode is not None:
            self._select_mode = select_mode
        else:
            self._select_mode = "single"

        # Handle MD's first
        for md in col_md:
            _addcol(md, allcols, precision=default_precision)

            if md.format == "perc_div":
                input_df[md.name] = input_df[md.name] / 100

        # Handle rest of columns not in MDs
        if is_tree:
            hidecols = [
                "index",
                pathcol,
                "date",
                "cast('t' as boolean)",
                "label",
                "full_label",
                "path_nodate",
                "path",
            ]
        else:
            hidecols = []
        for col in input_df.columns:
            if col not in col_md:
                md = ColMd(name=col, hidden=col in hidecols)
                _addcol(md, allcols, precision=default_precision)

        columns = list(allcols.values())

        splitfunc = f"""function(data) {{ 
                try {{
                    if (data === null || data.{pathcol} === null) 
                        return [];
                    else return data.{pathcol}.split('{pathdelim}');
                }} catch (error) {{
                    console.log(error);
                    console.log(data);
                    return [];
                }}}}
                
                """

        # Create a GridOptions object

        if num_toppinned_rows == 0:
            grid_data_df = input_df
            pinned_top_row_data_df = None

        else:
            grid_data_df = input_df[num_toppinned_rows:]
            pinned_top_row_data_df = input_df[0:num_toppinned_rows].to_dict(
                orient="records"
            )  # type: ignore

        args = {
            "license": license,
            "grid_data": grid_data_df,
            "grid_options": {
                "pinnedTopRowData": pinned_top_row_data_df,
                "columnDefs": columns,
                "defaultColDef": {
                    "sortable": "true",
                    "filter": "true",
                    "resizable": "true",
                },
                "enableRangeSelection": False,
                "rowSelection": select_mode,
                "treeData": is_tree,
                "getDataPath": splitfunc,
                "sideBar": {
                    "toolPanels": [
                        {
                            "id": "columns",
                            "labelDefault": "Columns",
                            "labelKey": "columns",
                            "iconKey": "columns",
                            "toolPanel": "agColumnsToolPanel",
                            "toolPanelParams": {
                                "suppressRowGroups": True,
                                "suppressValues": True,
                                "suppressPivots": True,
                                "suppressPivotMode": True,
                                "suppressColumnMove": True,
                            },
                        },
                        {
                            "id": "filters",
                            "labelDefault": "Filters",
                            "labelKey": "filters",
                            "iconKey": "filter",
                            "toolPanel": "agFiltersToolPanel",
                        },
                    ]
                },
                "autoGroupColumnDef": {
                    "headerName": "Path",
                    "pinned": "left",
                    "width": 300,
                    "suppressSizeToFit": True,
                    "suppressAutoSize": True,
                },
                **grid_options,
            },
            "quick_filter": True,
            "show_toggle_edit": False,
            "theme": "ag-theme-balham",
            "show_toggle_delete": False,
            "columns_fit": "auto",
            "index": sortcols,
            "keep_multiindex": showindex,
            "height": 500,  # px
            "width": "100%",
            "js_post_grid": [
                f"""
                let events={json.dumps(events)}
                //always include cell clicked so we can set the selected row
                if(!events){{
                    events=['cellClicked']
                }}
                else if(events.indexOf('cellClicked') < 0){{
                    events.push('cellClicked')
                }}

                events.forEach(eventName=>{{
                    console.log("Adding listener for",eventName)
                    view.gridOptions.api.addEventListener(eventName,(e)=>{{
                        console.log("Sending message for",eventName,e)
                        view.send({{
                            type: "gridEvent", 
                            event_type:eventName, 
                            col_clicked: e.column ? e.column.colId : undefined,
                            row_clicked: e.rowIndex,
                            value_clicked: e.value,
                            currentSelection:view.gridOptions.api.getSelectedNodes().map(n=>({{
                                key: n.key,
                                id: n.id,
                                data: n.data,
                                field:n.field,
                                leafData:n.allLeafChildren ? n.allLeafChildren.map(c=>c.data): undefined
                            }}))
                            }})
                        console.log("Sent message for",eventName)
                    }})
                    console.log("Added listener for",eventName)
                }})""",
                *js_post_grid,
            ],
            "js_pre_grid": [
                """
                console.log('******here***********',gridOptions);
                /*this isnt working, because the grid has no size 
                gridOptions.onFirstDataRendered = function onFirstDataRendered(params) {
                params.columnApi.autoSizeAllColumns();
                }*/
                let first=true;
                gridOptions.onGridSizeChanged=  function onFirstDataRendered(params) {
                
                if(first){
                      params.columnApi.autoSizeAllColumns();
                }
                }
                """,
                *js_pre_grid,
            ],
            **kargv,
        }

        super().__init__(*argv, **args)

        self.message_handlers = []

        def msg_rx(
            _, msg, buffers
        ):  # pragma: no cover  # bypassing because only reached from js
            try:
                self.process_message(msg)
            except Exception as e:
                logger.info("Unable to send msg")
                logger.exception(e)

        self.on_msg(msg_rx)

    def process_message(self, msg: Any):
        logger.debug("Widget got message: %s", msg)
        pMsg = msg

        # make sure its one of our events
        if "type" not in pMsg or pMsg["type"] != "gridEvent" and "event_type" in pMsg:
            return

        if self._select_mode is not None:
            if pMsg["event_type"] == "cellClicked" and "currentSelection" in pMsg:
                self.current_selection = pMsg["currentSelection"]

        logger.debug("Checking message handler for %s", len(self.message_handlers))
        for handler, handle_msg_type in self.message_handlers:
            logger.debug(
                "Checking message handler for %s and message type %s",
                handle_msg_type,
                pMsg["type"],
            )
            if handle_msg_type is None or pMsg["event_type"] == handle_msg_type:
                logger.debug("Calling handler")
                handler(pMsg)

    def on(
        self,
        msg_type: str,
        handler: Callable[[Dict], None],
    ):
        self.message_handlers.append((handler, msg_type))

    def remove_message_handler(self, handler: Callable[[Dict], None]):
        for handler_tuple in self.message_handlers:
            if handler_tuple[0] == handler:
                self.message_handlers.remove(handler_tuple)
                return True
        return False



import logging
from ipytree import Node, Tree

logger = logging.getLogger(__name__)


# def w_tree(root: str):
#    rootText = f'<span style="color: green;font-weight: italic;">{root}</span>'
#    tree = Tree(nodes=[Node(rootText)])
#    return tree


def w_tree_paths(paths: list[str], pathdelim):
    nodes = {}  # defaultdict(default_factor=list)
    rootnodes = []

    paths = sorted(paths)
    splitpaths = [p.split(pathdelim) for p in paths]
    pathsz = zip(paths, splitpaths)

    for ps, pl in pathsz:
        plEnd = pl[-1]
        nodeText = f'<span style="color: green;font-weight: italic;">{plEnd}</span>'
        n = Node(nodeText)

        n.path = ps  # type: ignore
        if len(pl) > 1:
            n.opened = False
        nodes[ps] = n
        if len(pl) > 1:
            parent_s = pathdelim.join(pl[0:-1])
            parent_node = nodes.get(parent_s)

            if parent_node is None:
                logger.warning(f"No parent node for {parent_s}, {ps}, {pl}")
            else:
                parent_node.add_node(n)

        else:
            rootnodes.append(n)

    def tree_value(self) -> list[Node]:
        selected = []
        for node in self.nodes:
            get_selected(node, selected)
        return selected

    def get_selected(t: Node, selected: list[Node]):
        if t.selected:
            selected.append(t.path)  # type: ignore
        for c in t.nodes:
            get_selected(c, selected)

    Tree.value = tree_value  # type: ignore

    tree = Tree()
    for n in rootnodes:
        tree.add_node(n)

    return tree



import ipyaggrid as ag
import pandas as pd
from .datagrid import ColMd, DataGrid
from typing import Optional, Callable, Literal


def display_ag(
    input_df: pd.DataFrame,
    is_tree: bool,
    pathcol: Optional[str] = "path",
    pathdelim="/",
    col_md: list[ColMd] = [],  # noqa: F821
    showindex: bool = False,
    action: Optional[Callable] = None,
    num_toppinned_rows: int = 0,
    grid_options: dict = {},
    flatten_columns: bool = True,
    default_precision: int = 2,
    select_mode: Optional[Literal["single", "multiple"]] = None,
) -> ag.Grid:

    grid = DataGrid(
        grid_data=input_df,
        is_tree=is_tree,
        pathcol=pathcol,
        pathdelim=pathdelim,
        col_md=col_md,
        showindex=showindex,
        num_toppinned_rows=num_toppinned_rows,
        grid_options=grid_options,
        default_precision=default_precision,
        flatten_columns=flatten_columns,
        select_mode=select_mode,
    )

    if action is not None:
        grid.on("cellClicked", action)

    grid.df = input_df  # type: ignore
    return grid



import enum
import io
import logging
from dataclasses import dataclass, field
from functools import partial
from inspect import signature
from typing import Annotated, Callable, ContextManager, Dict, List, Optional, Union

import ipytree
import ipyvuetify as v
import ipywidgets
import traitlets
from IPython.display import display
from ipyvuetify import Tab, TabItem
from ipywidgets import Widget

from . import BrowserTitle, datagrid, treew
from . import aggridhelper as g
from . import plotly_charts as pcharts

logger = logging.getLogger(__name__)

PATHDELIM = "~"


@dataclass
class UiWidget:
    w: Widget
    name: Optional[str]


class SelectTypes(enum.Enum):
    dropdown = 1
    combobox = 2
    slider = 3
    radio = 4


@dataclass
class UiPage:
    app: "UiModel"
    name: str

    widget: ipywidgets.Widget

    def clear(self):
        p = self.widget
        if p is not None:
            # TODO: Remove from self.widgets. Not critical, since
            # widgets are overwritten
            # by key value
            p.children = []  # type: ignore

    def add(self, elements: Union[UiWidget, List[UiWidget]]):
        self.app._add(target=self.name, elements=elements)

    def add_box(self, name: str, horiz: bool = False) -> "UiPage":
        """Adds either a horizontal or vertical box to contain the element.


        Args:
            page (str): App page to add this box to
            name (str): A globally unique name for the box widget, so it can \
                be later modified
            horiz (bool, optional): Horizontal or Vertical. Defaults to False (Vertical)

        Returns:
            _type_: _description_
        """
        if horiz:
            h = v.Html(tag="div", class_="d-flex flex-row", children=[])
        else:
            h = v.Html(tag="div", class_="d-flex flex-column", children=[])

        self.app.containers[name] = h

        self.add(
            elements=UiWidget(w=h, name=name),
        )

        return UiPage(self.app, name, h)  # type: ignore

    def add_separator(self, color: str = "gray", name: Optional[str] = None):
        w = v.Html(tag="hr", style_=f"border: none; border-top: 5px solid {color};")

        return self.add(
            elements=UiWidget(w=w, name=name),
        )

    def add_container(self, name: str) -> "UiPage":
        """Use add_box instead"""
        w = v.Container()
        self.app.containers[name] = w
        w = self.add(
            elements=UiWidget(w=w, name=name),
        )

        return UiPage(self.app, name, w)  # type: ignore

    def add_tree(
        self,
        paths: list[str],
        delim: str = PATHDELIM,
        name: Optional[str] = None,
    ):
        w = treew.w_tree_paths(paths=paths, pathdelim=delim if delim is not None else PATHDELIM)
        self.add(
            elements=UiWidget(w=w, name=name),
        )

    def add_textarea(
        self,
        label: str,
        name: Optional[str] = None,
        disabled: bool = False,
        value: str = "",
    ):
        w = v.Textarea(label=label, v_model=value, disabled=disabled)
        self.add(
            elements=UiWidget(w=w, name=name),
        )

    def add_textfield(
        self,
        label: Optional[str] = None,
        class_: Optional[str] = None,
        disabled: bool = False,
        value: str = "",
        solo=False,
        flat=False,
        name: Optional[str] = None,
    ):
        w = v.TextField(
            class_=class_,
            label=label,
            v_model=value,
            disabled=disabled,
            solo=solo,
            flat=flat,
        )
        return self.add(
            elements=UiWidget(w=w, name=name),
        )

    def add_progress(
        self,
        name: Optional[str] = None,
    ):
        w = v.ProgressLinear(class_="progress-bar", width=0, color="blue", indeterminate=True)
        w.hide()
        self.add(
            elements=UiWidget(w=w, name=name),
        )

    def add_textstatic(
        self,
        value: str = "",
        name: Optional[str] = None,
    ):
        w = v.CardText(children=value)
        return self.add(
            elements=UiWidget(w=w, name=name),
        )

    def add_textpre(
        self,
        value: str = "",
        name: Optional[str] = None,
    ):
        w = v.Html(tag="pre", children=[value], style_="max-height:80vh")

        return self.add(
            elements=UiWidget(w=w, name=name),
        )

    def add_output(
        self,
        max_outputs: Optional[int] = None,
        name: Optional[str] = None,
    ):
        w = ipywidgets.Output()
        w.add_class("messages_style")

        if max_outputs is not None:
            w.max_outputs = max_outputs  # type: ignore

        return self.add(
            elements=UiWidget(w=w, name=name),
        )

    def add_plt(
        self,
        plt,
        width=1024,
        height=1024,
        name: Optional[str] = None,
    ):
        buf = io.BytesIO()
        plt.fig.savefig(buf, format="png")
        image_widget = ipywidgets.widgets.Image(value=buf.getvalue(), format="png", width=width, height=height)

        return self.add(
            elements=UiWidget(w=image_widget, name=name),
        )

    def add_plotly_fig(
        self,
        fig,
        height=None,
        width=None,
        png=False,
        setcolors=True,
        name: Optional[str] = None,
    ):
        w = pcharts.create_widget(fig=fig, setcolors=setcolors, height=height, width=width, png=png)

        return self.add(
            elements=UiWidget(w=w, name=name),
        )

    def _make_action(self, caller_name: Optional[str], func: Callable) -> Callable:
        # Actions require 5 parameters. ALlow these parameters to be optional
        # component: str, action: str, args: str, app: nbapp.UiModel, caller: str

        # Make sure the function is a subset of expected keywords
        expected_keywords = {"component", "action", "args", "app", "caller"}
        keywords = set(signature(func).parameters.keys())
        unexpected_keywords = keywords - expected_keywords
        if unexpected_keywords:
            raise ValueError(
                f"{func.__name__} had an unexpected keywords: {unexpected_keywords}. Only \
                    {expected_keywords} were expected"
            )

        if len(keywords) != 5:
            logger.info(
                f"Action {func.__name__} does not have all five expected \
                    keywords, wrapping"
            )

            def func_wrapper(*args, **kwargs):
                # ipyvuetify passes 3 positional args: component, action, args

                new_kwargs = {}
                if "component" in keywords:
                    new_kwargs["component"] = args[0]
                if "action" in keywords:
                    new_kwargs["action"] = args[1]
                if "args" in keywords:
                    new_kwargs["args"] = args[2]
                if "app" in keywords:
                    new_kwargs["app"] = self.app
                if "caller" in keywords:
                    new_kwargs["caller"] = caller_name

                return func(**new_kwargs)

            return func_wrapper
        else:
            return partial(func, app=self.app, caller=caller_name)

    def add_select(
        self,
        label: str,
        disabled: bool = False,
        options: List = [],
        value=None,
        type: SelectTypes = SelectTypes.dropdown,
        multiple: bool = False,
        action: Optional[Callable] = None,
        horiz: bool = False,
        name: Optional[str] = None,
    ):
        if type == SelectTypes.dropdown:
            w = v.Select(
                label=label,
                disabled=disabled,
                items=options,
                multiple=multiple,
                v_model=value,
            )
        elif type == SelectTypes.combobox:
            w = v.Combobox(
                label=label,
                disabled=disabled,
                items=options,
                multiple=multiple,
                v_model=value,
            )
        elif type == SelectTypes.slider:
            w = v.Slider(
                label=label,
                disabled=disabled,
                v_model=value if value is not None else options[0],
                min=options[0],
                max=options[-1],
                step=1,
                thumb_label=True,
            )
        elif type == SelectTypes.radio:
            children = [v.Radio(label=str(o), value=str(o)) for o in options]
            w = v.RadioGroup(label=label, children=children, v_model=None, row=horiz)
        else:
            raise ValueError(f"Unexpected type: {type}")

        if action is not None:
            action = self._make_action(name, action)
            w.on_event("change", action)

        return self.add(
            elements=UiWidget(w=w, name=name),
        )

    def add_button(
        self,
        label: str,
        action: Callable[..., None],
        disabled: bool = False,
        status: bool = False,
        name: Optional[str] = None,
    ):
        """

        Args:
            page (str): The tab (page) or the container to add the button to.
            name (str): Internal name used to reference the button widget
            action (Callable): Callable that is executed when button is clicked. Must \
                take four parameters: component, action, args, app, and caller.
            disabled (bool, optional): A disabled button cannot be clicked. \
                Defaults to False.
            status (bool, optional): Adds a status field to the right of the button. \
                Useful for showing feedback to user of application process. \
                    Defaults to False.
            label (Optional[str], optional): Label to show to side of button. \
                Defaults to None.

        Returns:
            _type_: _description_
        """

        if name is None:  # label must be set, but default to the name
            name = label

        if status:
            status_container = name + ".box"
            status_box = self.add_box(name=status_container, horiz=False)
        else:
            status_box = None

        b = v.Btn(children=[label], disabled=disabled, outlined=True)

        if action is not None:
            action = self._make_action(name, action)
            b.on_event("click", action)

        if status_box:
            o = status_box.add(elements=UiWidget(w=b, name=name))
        else:
            o = self.add(elements=UiWidget(w=b, name=name))

        if status:
            self.add_textfield(
                name=name + ".status",
                label=None,
                disabled=True,
                value="Not Run",
                solo=True,
                flat=True,
                class_="for-progress",
            )

            progress_container = name + ".box.progress"
            box = self.add_box(name=progress_container, horiz=False)

            box.add_progress(name=name + ".status_bar")
        return o

    def add_df(
        self,
        df,
        tree: bool = False,
        pathcol: Optional[str] = None,
        col_md: List[g.ColMd] = [],
        showindex: bool = False,
        action: Optional[Callable] = None,
        num_toppinned_rows: int = 0,
        table_viewer: bool = False,
        flatten_columns: bool = True,
        pathdelim: str = PATHDELIM,
        precision: int = 2,
        grid_options: dict = {},
        multiselect: bool = False,
        name: Optional[str] = None,
    ):
        if table_viewer is True:
            grid_options["autoGroupColumnDef"] = {
                "cellRendererParams": {
                    "suppressCount": True,
                    "suppressGroupCheckbox": """function(params) {
                        return params.node.allLeafChildren.length === 1;
                    }""",
                },
            }

            grid_options["groupDefaultExpanded"] = -1

        if multiselect:
            select_mode = "multiple"
        else:
            select_mode = None

        if tree and pathcol not in df.columns:
            raise ValueError("If tree, then pathcol must be in df.columns")

        w = g.display_ag(
            df,
            tree,
            pathcol=pathcol,
            pathdelim=pathdelim,
            showindex=showindex,
            col_md=col_md,
            action=action,
            num_toppinned_rows=num_toppinned_rows,
            grid_options=grid_options,
            flatten_columns=flatten_columns,
            default_precision=precision,
            select_mode=select_mode,
        )

        return self.add(
            elements=UiWidget(w=w, name=name),
        )

    def add_html(
        self,
        html: str,
        name: Optional[str],
    ):
        ho = HTMLOutput()
        ho.html = html
        w = UiWidget(name=name, w=ho)
        self.add(w)


@dataclass
class UiModel:
    pages: List[str]

    log_page: Optional[str] = None

    _page_widgets: Dict[str, UiPage] = field(default_factory=dict)

    messages: Annotated[ContextManager[str], field(init=False)] = field(init=False)

    # Containers are used for replacing existing elements and are optional
    containers: Dict[str, ipywidgets.Widget] = field(default_factory=dict)

    # Every UI widget is stored by name (globally in the app). This is used
    # to retrieve the current value of individual widgets.
    widgets: Dict[str, UiWidget] = field(default_factory=dict)

    title: Optional[str] = None

    def get_valuestr(self, name):
        return str(self.get_values(name))  # type: ignore

    def get_values(self, name):
        w = self.widgets[name].w

        if isinstance(w, datagrid.DataGrid):
            return w.current_selection
        if isinstance(w, ipytree.tree.Tree):
            return w.value()  # type: ignore
        else:
            return self.widgets[name].w.v_model  # type: ignore

    def _add(self, target: str, elements: Union[UiWidget, List[UiWidget]]):
        """Target can be either a Page name or a Container name"""
        if isinstance(elements, UiWidget):
            elements = [elements]

        if target in self.containers:
            logger.debug("Adding to container")
            widgets = [e.w for e in elements]
            self.containers[target].children = (  # type: ignore
                *self.containers[target].children,  # type: ignore
                *widgets,
            )
            for e in elements:
                if e.name is not None:
                    self.widgets[e.name] = e
        else:
            logger.debug("Adding to page")

            page = self._page_widgets[target].widget  # type: ignore

            children = page.children  # type: ignore
            newchildren = []
            for e in elements:
                if e.name is not None:
                    self.widgets[e.name] = e

                newchildren.append(e.w)

            page.children = (*children, *newchildren)  # type: ignore

    def get_container(self, name: str) -> "UiPage":
        w = self.containers[name]
        return UiPage(self, name, w)  # type: ignore

    def get_page(self, title: Union[str, int]) -> UiPage:
        """

        Args:
            title (Union[str, int]): If int, returns the page at the offset given from \
                self.pages. If str, returns the page by name

        Raises:
            ValueError: An error if the input is not a Page

        """
        if isinstance(title, int):
            title = self.pages[title]

        if title in self._page_widgets:
            return self._page_widgets[title]
        else:
            raise ValueError(f"{title} is not a page")

    def set_value(self, name, value):
        self.widgets[name].w.v_model = value  # type: ignore

    def clear_messages(self):
        self.messages.clear_output()  # type: ignore

    def clear_container(self, id: str):
        w = self.widgets[id].w
        w.children = []  # type: ignore

    def update_status(self, name: str, message: str, running: bool = True):
        self.set_value(f"{name}.status", message)
        if running:
            self.start_progress(name)
        else:
            self.end_progress(name)

    def start_progress(self, name):
        self.widgets[f"{name}.status_bar"].w.show()  # type: ignore

    def end_progress(self, name):
        self.widgets[f"{name}.status_bar"].w.hide()  # type: ignore


@dataclass
class TabbedUiModel(UiModel):
    _tabWidget: v.Tabs = None  # type: ignore  - initialized in post_init
    _headerWidget: Optional[ipywidgets.Widget] = None
    _footerWidget: Optional[ipywidgets.Widget] = None
    log_footer: Optional[str] = "Messages"

    # the pages to put in the header
    headers: Optional[List[str]] = None

    def __post_init__(self):
        children = []
        if self.headers is not None:
            for header in self.headers:
                self._page_widgets[header] = UiPage(app=self, name=header, widget=v.ExpansionPanelContent(children=[]))
                self._headerWidget = v.ExpansionPanels(
                    children=[
                        v.ExpansionPanel(
                            children=[
                                v.ExpansionPanelHeader(children=[header]),
                                self._page_widgets[header].widget,
                            ]
                        )
                    ],
                    v_model=[0],
                    multiple=True,
                )

        for tname in self.pages:
            t = Tab(children=[tname])
            ti = TabItem(children=[], style_="padding-left: 20px; padding-right: 20px;")
            children.append(t)
            children.append(ti)

            self._page_widgets[tname] = UiPage(app=self, name=tname, widget=ti)

        self._tabWidget = v.Tabs(v_model=[0], children=children)

        log_page_name = self.log_page

        if self.log_footer is not None:
            log_page_name = self.log_footer
            self._page_widgets[log_page_name] = UiPage(
                app=self,
                name=log_page_name,
                widget=v.ExpansionPanelContent(children=[]),
            )
            self._footerWidget = v.ExpansionPanels(
                children=[
                    v.ExpansionPanel(
                        children=[
                            v.ExpansionPanelHeader(children=[log_page_name]),
                            self._page_widgets[log_page_name].widget,
                        ]
                    )
                ]
            )
        if log_page_name is not None:
            self.get_page(log_page_name).add_output(name="m.ta", max_outputs=100)
            self._tabWidget = v.Tabs(v_model=[0], children=children)
            self.messages = self.widgets["m.ta"].w  # type: ignore

        if self.title is not None:
            BrowserTitle(self.title)

    def add_page(self, title: str, children: List = [], selected_tab=True) -> ipywidgets.Widget:
        self.pages.append(title)

        tab_children = [*self._tabWidget.children]

        t = Tab(children=[title])
        ti = TabItem(children=children, style_="padding-left: 20px; padding-right: 20px;")
        tab_children.append(t)
        tab_children.append(ti)
        self._page_widgets[title] = UiPage(app=self, name=title, widget=ti)

        self._tabWidget.children = tab_children
        if selected_tab:
            self._tabWidget.v_model = len(tab_children) / 2 - 1

        return self._page_widgets[title]  # type: ignore

    def remove_page(self, title: str):
        if title in self.pages:
            index = list(self.pages).index(title)
            self.pages.remove(title)

            tab_children = [*self._tabWidget.children]
            del tab_children[index * 2 : index * 2 + 2]

            self._page_widgets[title] = None  # type: ignore
            self._tabWidget.children = tab_children

            if self._tabWidget.v_model == index:
                self._tabWidget.v_model = index - 1

    def open_tab(self, name: str):
        if name in self.pages:
            self._tabWidget.v_model = list(self.pages).index(name)

    def toggle_tab(self, name: str, disabled: bool):
        if name in self.pages:
            index = list(self.pages).index(name)
            index *= 2
            self._tabWidget.children[index].disabled = disabled
        if self.headers is not None and name in self.headers:
            index = list(self.headers).index(name)
            self._headerWidget.children[index].disabled = disabled  # type: ignore

    def get_children(self):
        children = []
        if self._headerWidget is not None:
            children.append(self._headerWidget)
        if self._tabWidget is not None:
            children.append(self._tabWidget)
        if self._footerWidget:
            children.append(self._footerWidget)
        return children

    def display(self):
        display(ThemeFixer())
        return v.Html(
            tag="div",
            children=[
                v.Html(
                    tag="style",
                    children=[
                        """
    .vuetify-styles code, .vuetify-styles kbd, .vuetify-styles samp{
        color: black !important
    }

    // .vuetify-styles pre

    .v-tabs div{
        transition: none !important;
    }

    .for-progress .v-text-field__details{
        display: none !important;
    }

    .for-progress .v-input__control{
        min-height: 0px !important;
    }

    .for-progress input{
        text-align: center;
    }

    .progress-bar{
        margin-left: 10px
    }

    .vuetify-styles .v-container{
        min-width: 80vw
    }

    .ag-header {
        position: relative;
    }

    .messages_style {
        color: pink; !important;
    }


                """
                    ],
                ),
                v.Container(children=self.get_children()),
            ],
        )


class HTMLOutput(v.VuetifyTemplate):
    html = traitlets.Unicode(default_value=None, allow_none=True).tag(sync=True)

    @traitlets.default("template")
    def _template(self):
        return """
            <template>
                <div v-html="html">
                </div>
            </template>
        """


class ThemeFixer:
    def _repr_html_(self) -> str:
        return """<script>
                    if (window.location.href.indexOf('voila') >= 0){
                        const l=document.createElement('link');
                        l.setAttribute('rel','stylesheet');
                        l.setAttribute('type','text/css');
                        l.setAttribute('href',`${window.location.href.split('/').slice(0,7).join('/')}/static/theme-light.css`);
                        document.body.appendChild(l);
                        document.body.classList.remove('theme-dark')
                        document.body.classList.add('theme-light')
                    }
                </script>
                """
